package ds;
import org.garret.perst.IInputStream;
import org.garret.perst.IOutputStream;
import org.garret.perst.Persistent;
import org.garret.perst.Storage;

/** One dictionary index entry.
  * Dictionary index entries are keyed on the keywords and
  * contain an array of line numbers in the dictionary file for the
  * explanation of the keyword.
  * @author Daniel Stoinski
  * @version $Revision: 28 $
  */
public class DictIndexEntry extends Persistent {

  /** Empty index keyword.
    */
  private static final String EMPTY_KEYWORD = "";

  /** Empty index entry lines.
    */
  private static final int[] EMPTY_LINES = new int[0];

  /** The keyword of the dictionary index entry.
    * The keyword may be for example the English word in an
    * English-German dictionary.
    */
  private String m_keyword;

  /** The arrasy of lines for the dictionary entries.
    */
  private int[] m_lines;

  /** Initializes an empty dictionary index entry.
    */
  public DictIndexEntry() {
    super();
    this.m_keyword = DictIndexEntry.EMPTY_KEYWORD;
    this.m_lines = DictIndexEntry.EMPTY_LINES;
  } // DictIndexEntry()

  /** Initializes adisctionary index entry for the given keyword and lines.
    * @param aKeyWord  the keyword
    * @param lines     the lines in the dictionary file with the explanations
    *                  of the keyword.
    */
  public DictIndexEntry(String aKeyWord, int[] lines) {
    super();
    this.m_keyword = aKeyWord == null? DictIndexEntry.EMPTY_KEYWORD : aKeyWord;
    this.m_lines = lines == null? DictIndexEntry.EMPTY_LINES : lines;
  } // DictIndexEntry()

   
  /** Initializes adisctionary index entry for the given keyword and lines.
    * @param aStorage  the Perst storage
    * @param aKeyWord  the keyword
    * @param lines     the lines in the dictionary file with the explanations
    *                  of the keyword.
    */
  public DictIndexEntry(Storage aStorage, String aKeyWord, int[] lines) {
    super(aStorage);
    this.m_keyword = aKeyWord == null? DictIndexEntry.EMPTY_KEYWORD : aKeyWord;
    this.m_lines = lines == null? DictIndexEntry.EMPTY_LINES : lines;
  } // DictIndexEntry()

  /** Returns the keyword for the dictionary index entry.
    * @return the keyword
    */
  public String getKeyWord() {
    return this.m_keyword;
  } // getKeyWord()

  /** Returns an array of line numbers with the explanations for the keyword.
    * @return array of line numbers.
    */
  public int[] getLines() {
    return this.m_lines;
  } // getLines()

  /** Returns the string representation of this object.
    * String with the following format:
    * keyword numlines line(0) ... line(numlines-1)
    * @return this object as string.
    */
  public String toString() {
    StringBuffer sb;
    int          i;

    sb = new StringBuffer();
    sb.append(this.m_keyword).append(' ').append(this.m_lines.length);
    for(i = 0; i < this.m_lines.length; i++)
      sb.append(' ').append(this.m_lines[i]);
    return sb.toString();
  } // toString()

  /** Deserializes the object.
    * Reads the object contents from the given input stream. Keyword, number
    * of lines, line numbers.
    * @param in  the input stream fo read the object from.
    */
  public void readObject(IInputStream in) {
    int i, n;

    this.m_keyword = in.readString();
    n = in.readInt();
    this.m_lines = new int[n];
    for(i = 0; i < n; i++) this.m_lines[i] = in.readInt();
  } // readObject()

  /** Serializes the object.
    * Writes the contents of the object to the given output stream in the same
    * order, as expected in readObject()
    * @param out the stream to write the object to.
    */
  public void writeObject(IOutputStream out) {
    int i;

    out.writeString(this.m_keyword);
    out.writeInt(this.m_lines.length);
    for(i = 0; i < this.m_lines.length; i++) out.writeInt(this.m_lines[i]);
  } // writeObject()
} // class DictIndexEntry

