package ds;
import org.garret.perst.Index;
import org.garret.perst.Key;
import org.garret.perst.Storage;
import org.garret.perst.StorageFactory;
import javax.microedition.io.Connector;
import java.io.InputStream;

/** Imports the dictionary generated by prepdict into the db.
  * @author Daniel Stoinski
  * @version $Revision: 31 $
  */
public class Import extends Thread {

  /** Separators for tokens in readTokens.
    * @see #readTokens
    */
  private static final String TOKENSEPARATORS = " \n\r\t";

  /** Line separators used for reading dictionary entries.
    */
  private static final String LINESEPARATORS = "\n\r";

  /** Encoding of the input file.
    */
  private static final String ENCODING = "Cp1252";

  /** Length of help buffers used in order to parse input lines.
    */
  private static final int BUFLEN = 4096;

  /** The perst page pool size.
    */
  public static final int PAGE_POOL_SIZE = 65536;

  /** State: doing nothing.
    * @see #newToken
    */
  private static final int STATE_NOTHING = 0;

  /** State: read number of index keywords.
    * @see #newToken
    */
  private static final int STATE_RD_NUM_KEYWORDS = 1;

  /** State: read number of dictionary entries.
    * @see #newToken
    */
  private static final int STATE_RD_NUM_ENTRIES = 2;

  /** State: read index keyword.
    * @see #newToken
    */
  private static final int STATE_RD_KEYWORD = 3;

  /** State: read number of line numbers for the index keyword.
    * @see #newToken
    */
  private static final int STATE_RD_NUM_LINENO = 4;

  /** State: read the line number itself.
    * @see #newToken
    */
  private static final int STATE_RD_LINENO = 5;

  /** State: read the dictionary entry.
    * @see #newToken
    */
  private static final int STATE_RD_ENTRY = 6;

  /** Handle of the Perst database.
    */
  private Storage m_db;

  /** The root of the created dictionary database.
    */
  private DictRoot m_root;
  /** Eception occured during the processing.
    */
  private Exception m_excp;

  /** Form for showing progress and errors.
    */
  private IDictForm m_form;

  /** Name of the dictionary index file to import.
    */
  private String m_idxfile;

  /** Name of the dictionary file to import.
    */
  private String m_dictfile;

  /** Name of the database to create.
    */
  private String m_dbname;

  /** Current index keyword processed in newToken.
    */
  private String m_keyword;

  /** Current array of line numbers for the keyword processed in newToken.
    */
  private int[] m_linenos;

  /** Index of the current line number in m_linenos processed in newToken.
    */
  private int m_linenoidx;

  /** Current status of the object for the processing in newToken.
    */
  private int m_status;

  /** Number of index keywords.
    * Read from the dictionary index.
    */
  private int m_n_keywords;

  /** Current keyword added to the DB
    */
  private int m_curkeyword;

  /** Number of dictionary entries.
    * Also read from the dictionary index.
    */
  private int m_n_entries;

  /** State of the objecto to consider in newToken.
    * @see #newToken
    */
  private short m_state;

  /** Break the import if goon becomes false.
    */
  private boolean m_goon;

  /** Imports the dictionary to the given database.
    * @param aDictFileName    name of the dictionary file as obtained from 
    *                         dict.cc.
    * @param anIdxFileName    name of the index file as created with prepdict.
    * @param aDBName          name of the Perst database into which to
    *                         store the dictionary.
    */
  public Import(IDictForm form,
                String    aDictFileName,
                String    anIdxFileName,
                String    aDBName) {
    this.m_db = null;
    this.m_root = null;
    this.m_excp = null;
    this.m_form = form;
    this.m_keyword = null;
    this.m_linenos = null;
    this.m_linenoidx = 0;
    this.m_status = Import.STATE_NOTHING;
    this.m_n_keywords = 0;
    this.m_curkeyword = 0;
    this.m_n_entries = 0;
    this.m_goon = false;
    this.m_idxfile = anIdxFileName;
    this.m_dictfile = aDictFileName;
    this.m_dbname = aDBName;
  } // Import()

  /** Creates Perst database.
    */
  private void openDB() {
    if (this.m_db == null) {
      this.m_db = StorageFactory.getInstance().createStorage();
      this.m_db.open(this.m_dbname, Import.PAGE_POOL_SIZE);
    }
    this.m_root = new DictRoot(this.m_db);
    this.m_db.setRoot(this.m_root);
  } // openDB()

  /** Closes Perst database.
    */
  private void closeDB() {
    if (this.m_db != null) {
      try {
        this.m_db.commit();
      }
      catch(Exception excp) {
        if (this.m_excp == null) this.m_excp = excp;
      }
      try {
        this.m_db.close();
      }
      catch(Exception excp) {
        if (this.m_excp == null) this.m_excp = excp;
      }
    }
    this.m_root = null;
    this.m_db = null;
  } // closeDB()

  /** Stops the thread.
    */
  public void stop() {
    this.m_goon = false;
    this.interrupt();
  } // stop()

  /** Adds a new dictionary index entry to the database.
    * Creates a new dictionary index entry from m_keyword and m_linenos
    * and adds it to the database.
    */
  private void addDictIndexEntry() {
    DictIndexEntry die;
    Index          idx;

    if (this.m_goon && this.m_root != null &&
        this.m_keyword != null && this.m_linenos != null &&
        this.m_keyword.length() > 0 && this.m_linenos.length > 0) {
      die = new DictIndexEntry(this.m_keyword, this.m_linenos);
      idx = this.m_root.getKwIdx();
      idx.put(this.m_keyword.toUpperCase().trim(), die);
      if (this.m_form != null && this.m_n_keywords != 0) {
        this.m_form.progress(this.m_curkeyword * 100 / this.m_n_keywords);
        this.m_curkeyword++;
      }
    }
  } // addDictIndexEntry()

  /** Adds the given dictionary entry to the database.
    * @param anEntry  the dictionary line add
    * @param aLineNo  the line number of the dictionary entry (index).
    */
  private void addDictEntry(String anEntry, int aLineNo) {
    DictEntry entry;
    Index     idx;

    if (this.m_goon && this.m_root != null &&
        anEntry != null && anEntry.length() > 0 && aLineNo >= 0) {
      entry = new DictEntry(aLineNo, anEntry);
      idx = this.m_root.getDictIdx();
      idx.put(new Key(aLineNo), entry);
      if (this.m_form != null && this.m_n_entries != 0) {
        this.m_form.progress(aLineNo * 100 / this.m_n_entries);
      }
    }
  } // addDictEntry

  /** Handles tokes read by readTokes while parsing the index file.
    * This function is called for every token from the dictionary index file
    * or for every line in the dictionary file. What do the method particulaty
    * have to do at the moment is stored in the variable m_status.
    * The following status codes are recognized by the method:
    * <TABLE BORDER>
    *  <TR>
    *   <TD>STATE_NOTHING</TD>
    *   <TD>
    *    The object was just about to start reading the dictionary index
    *    file and the dictionary file. Read number of keywords in the 
    *    dictionary index file and go to the state STATE_RD_NUM_KEYWORDS.
    *   </TD>
    *  </TR>
    *  <TR>
    *   <TD>STATE_RD_NUM_KEYWORDS</TD>
    *   <TD>
    *    The number of keywords in the dictionary index file has been 
    *    successfully read. Read the number of dictionary entries
    *    and switch to the state STATE_RD_NUM_ENTRIES.
    *   </TD>
    *  </TR>
    *  <TR>
    *   <TD>STATE_RD_NUM_ENTRIES</TD>
    *   <TD>
    *    The number of dictionary entries has been successfully read.
    *    Read the first dictionary index keyword and switch to the state
    *    STATE_RD_KEYWORD.
    *   </TD>
    *  </TR>
    *  <TR>
    *   <TD>STATE_RD_KEYWORD</TD>
    *   <TD>
    *    The dictionary index keyword has been successfully read and temporyrly
    *    stored in m_keyword. Read the number of line
    *    numbers of entries for the keyword, create the temporary buffer
    *    m_linenos and go to the state STATE_RD_NUM_LINENO.
    *   </TD>
    *  </TR>
    *  <TR>
    *   <TD>STATE_RD_NUM_LINENO</TD>
    *   <TD>
    *    The number of line numbers for the dictionary index keyword has been
    *    successfully read and the buffer for the line numbers successfully
    *    created in m_linenos. If the number of line numbers is 0, then
    *    read the next keyword going to STATE_RD_KEYWORD. Else read the line
    *    numbers going to STATE_RD_LINENO.
    *   </TD>
    *  </TR>
    *  <TR>
    *   <TD>STATE_RD_LINENO</TD>
    *   <TD>
    *    Next line number of the dictionary entry of the current keyword
    *    has been successfully read and stored in m_linenos[m_linenoidx].
    *    If all line numbers for the keyword have been successully read,
    *    then build a dictionary index entry object DictIndexEntry, store
    *    it in the database and read the next keyword going to
    *    STATE_RD_KEYWORD. Else stay in this state in order to be able to
    *    read the next line number.
    *   </TD>
    *  </TR>
    *  <TR>
    *   <TD>STATE_RD_ENTRY</TD>
    *   <TD>
    *    The only state for reading dictionary entries rather than the
    *    dictionary index entries. Store the whole dictionary entry
    *    in the database and stay in this state.
    *   </TD>
    *  </TR>
    * </TABLE>
    * @param aToken  the token read in readTokens()
    * @param aSeqNum the sequential number of the token or of the line.
    * @see #readTokens
    */
  private void newToken(String aToken, int aSeqNum) {
    int            n;

    switch(this.m_status) {
      case Import.STATE_NOTHING :
        this.m_n_keywords = Integer.parseInt(aToken);
        this.m_status = STATE_RD_NUM_KEYWORDS;
        break;
      case Import.STATE_RD_NUM_KEYWORDS :
        this.m_n_entries = Integer.parseInt(aToken);
        this.m_status = Import.STATE_RD_NUM_ENTRIES;
        break;
      case Import.STATE_RD_NUM_ENTRIES :
        this.m_keyword = aToken;
        this.m_linenos = null;
        this.m_linenoidx = 0;
        this.m_status = Import.STATE_RD_KEYWORD;
        break;
      case Import.STATE_RD_KEYWORD :
        n = Integer.parseInt(aToken);
        this.m_linenos = new int[n];
        this.m_linenoidx = 0;
        this.m_status = Import.STATE_RD_NUM_LINENO;
        break;
      case Import.STATE_RD_NUM_LINENO :
      case Import.STATE_RD_LINENO :
        if (this.m_linenoidx < this.m_linenos.length) {
          n = Integer.parseInt(aToken);
          this.m_linenos[this.m_linenoidx] = n;
          this.m_linenoidx++;
          if (this.m_linenoidx == this.m_linenos.length) {
            this.addDictIndexEntry();
            this.m_status = Import.STATE_RD_NUM_ENTRIES;
          }
          else {
            this.m_status = Import.STATE_RD_LINENO;
          }
        }
        else {
          this.m_status = Import.STATE_RD_NUM_ENTRIES;
          this.newToken(aToken, aSeqNum);
        }
        break;
      case Import.STATE_RD_ENTRY :
        this.addDictEntry(aToken, aSeqNum + 1);
        break;
    }
  } // nowToken()

  /** Reads tokens from the input stream.
    * For each token the method calls newToken().
    * @param is               the input stream
    * @param tokenSeparators  characters separating tokens
    * @see #newToken
    */ 
  private void readTokens(InputStream is, String tokenSeparators) {
    String s;
    byte   b;
    byte[] bin, bstr;
    int    i, len, actidx, token_count;

    try {
      if (is != null) {
        bin = new byte[Import.BUFLEN];
        bstr = new byte[Import.BUFLEN];
        actidx = 0;
        token_count = 0;
        while ((len = is.read(bin)) != -1) {
          for(i = 0; i < len; i++) {
            b = bin[i];
            if (tokenSeparators.indexOf(b) == -1) {
              if (actidx < Import.BUFLEN) {
                bstr[actidx] = b;
                actidx++;
              }
            }
            else {
              if (actidx > 0) {
                s = new String(bstr, 0, actidx, Import.ENCODING);
                this.newToken(s, token_count);
                token_count++;
              }
              actidx = 0;
            }
          } // for i
        } // while
        if (actidx > 0) {
          s = new String(bstr, 0, actidx, Import.ENCODING);
          this.newToken(s, token_count);
        }
      } // if (is != null)
    }
    catch(Exception excp) {
      if (this.m_excp != null) this.m_excp = excp;
    }
  } // readTokens()

  /** The thread method makes the whole job.
    */
  public void run() {
    InputStream     is;

    this.m_keyword = null;
    this.m_linenos = null;
    this.m_linenoidx = 0;
    this.m_n_keywords = 0;
    this.m_curkeyword = 0;
    this.m_n_entries = 0;
    this.m_goon = true;
    is = null;
    try {
      this.openDB();
      is = Connector.openInputStream(this.m_idxfile);
      this.m_status = Import.STATE_NOTHING;
      this.readTokens(is, Import.TOKENSEPARATORS);
      is.close();
      this.m_status = Import.STATE_RD_ENTRY;
      is = Connector.openInputStream(this.m_dictfile);
      this.readTokens(is, Import.LINESEPARATORS);
    }
    catch(Exception excp) {
      if (this.m_excp != null) this.m_excp = excp;
    }
    this.m_goon = false;
    if (is != null) { try { is.close(); } catch(Exception excp) {} }
    this.closeDB();
    if (this.m_excp != null && this.m_form != null) {
      this.m_form.message(m_excp);
    }
  } // run()
} // class ImportIndexEntries

